/* $Id: complex.h,v 1.5 2004/03/04 16:41:57 scorzato Exp $ */

#ifndef _COMPLEX_H
#define _COMPLEX_H

#include <math.h>
#include "sse-extension/sse.h"

typedef struct
{
   double re,im;
} complex;

/* x = x+conj(z)*w */
#ifdef SSE3
#define _add_assign_complex_conj(x,z,w) \
__asm__ __volatile__ ("" \
                      : \
                      : \
                      "");

#else
/* To be compatible to standard C complex.h */
#define cimag(x) (x).re
#define creal(x) (x).im

#define _add_assign_complex_conj(x,z,w) \
   (x).re += (z).re*(w).re+(z).im*(w).im; \
   (x).im += (-(w).re*(z).im+(w).im*(z).re)
#endif

/* x = x + z*w  */
#define _add_assign_complex(x,z,w) \
   (x).re += (z).re*(w).re-(z).im*(w).im; \
   (x).im += (w).re*(z).im+(w).im*(z).re

/* x = x- conj(z)*w */
#define _diff_assign_complex_conj(x,z,w) \
   (x).re -= (z).re*(w).re+(z).im*(w).im; \
   (x).im -= -(w).re*(z).im+(w).im*(z).re

/* x = x - z*w  */
#define _diff_assign_complex(x,z,w) \
   (x).re -= (z).re*(w).re-(z).im*(w).im; \
   (x).im -= (w).re*(z).im+(w).im*(z).re

/* x = conj(z)*w  */
#define _mult_assign_complex_conj(x,z,w) \
   (x).re = (z).re*(w).re+(z).im*(w).im; \
   (x).im = -(w).re*(z).im+(w).im*(z).re

/* x = z*w */
#define _mult_assign_complex(x,z,w) \
   (x).re = (z).re*(w).re-(z).im*(w).im; \
   (x).im = (w).re*(z).im+(w).im*(z).re

/*  z = z - w */
#ifdef SSE3
#define _diff_complex(z,w) \
__asm__ __volatile__ ("movapd %0, %%xmm0 \n\t" \
                      "movapd %1, %%xmm1 \n\t" \
                      "subpd %%xmm1, %%xmm0 \n\t" \
                      "movapd %%xmm0, %0 \n\t" \
                      : \
                      : \
                      "m" (z), \
                      "m" (w))
#else
#define _diff_complex(z,w) \
   (z).re -= (w).re; \
   (z).im -= (w).im
#endif

#ifdef SSE3
#define _add_complex(z,w) \
__asm__ __volatile__ ("movapd %0, %%xmm0 \n\t" \
                      "movapd %1, %%xmm1 \n\t" \
                      "addpd %%xmm1, %%xmm0 \n\t" \
                      "movapd %%xmm0, %0 \n\t" \
                      : \
                      : \
                      "m" (z), \
                      "m" (w))
#else
#define _add_complex(z,w) \
   (z).re += (w).re; \
   (z).im += (w).im
#endif

/* z = r * w, r real */
#ifdef SSE3
#define _mult_real(z,w,r) \
__asm__ __volatile__ ("movsd %2, %%xmm0 \n\t" \
                      "unpcklpd %%xmm0, %%xmm0 \n\t" \
                      "movapd %1, %%xmm1 \n\t" \
                      "mulpd %%xmm0, %%xmm1 \n\t" \
                      "movapd %%xmm1, %0 \n\t" \
                      : \
                      : \
                      "m" (z), \
                      "m" (w), \
                      "m" (r))
#else
#define _mult_real(z,w,r) \
   (z).re = r * (w).re; \
   (z).im = r * (w).im 
#endif

/* ||z||^2 */
#define _complex_square_norm(z) \
   (z).re*(z).re+(z).im*(z).im

/* ||z|| */
#define _complex_norm(z) \
   sqrt(_complex_square_norm(z)) 

/* z = 0 */
#define _complex_zero(z) \
   (z).re=0.;(z).im=0.

/* z = 1 */
#define _complex_one(z) \
   (z).re=1.;(z).im=0.

/* |z.re|+|z.im| */
#define _complex_1norm(z) \
   fabs((z).re) + fabs((z).im)

/* z = r + i*s */
#define _complex_set(z,r,s) \
   (z).re=r;(z).im=s

/* x = z / w */
#define _div_complex(x,z,w) \
   (x).re = ((z).re*(w).re+(z).im*(w).im)/((w).re*(w).re+(w).im*(w).im); \
   (x).im = ((z).im*(w).re-(z).re*(w).im)/((w).re*(w).re+(w).im*(w).im)

/* z = w / r, r real */
#define _div_real(z,w,r) \
   (z).re = (w).re / r; \
   (z).im = (w).im / r

/* z = conj(w) */
#define _complex_conj(z,w) \
   (z).re = (w).re ; \
   (z).im = -(w).im

/* z = -w */
#define _complex_chgsig(z,w) \
   (z).re = -(w).re ; \
   (z).im = -(w).im

#endif





























