/***************************************************************************
! initialize.c
!
! init routines adapted from GWC-package provided by Carsten Urbach
!
! author: Andre Sternbeck <andre.sternbeck@physik.hu-berlin.de
****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#ifdef MPI
#include <mpi.h>
#endif
# include "global.h"
# include "su3.h"

int    *iup, *idn, *ipt, **ipt_, ***ipt__;
su3    *gauge_field;
static int ini=0;

void initialize(void)
{
  int i;

  if(ini == 0){
    ini=1;

    g_idn = calloc(VOLUMEPLUSRAND, sizeof(int*));
    if(errno == ENOMEM) die("Couldn't get memory for g_idn\n",errno);
    g_iup = calloc(VOLUMEPLUSRAND, sizeof(int*));
    if(errno == ENOMEM) die("Couldn't get memory for g_iup\n",errno);
    g_gauge_field = calloc(VOLUMEPLUSRAND, sizeof(su3*));
    if(errno == ENOMEM) die("Couldn't get memory for g_gauge_field\n",errno);
    ipt_ = calloc(T*LX*LY, sizeof(int*));
    if(errno == ENOMEM) die("Couldn't get memory for ipt_\n",errno);
    ipt__ = calloc (T*LX, sizeof(int*));
    if(errno == ENOMEM) die("Couldn't get memory for ipt__\n",errno);
    g_ipt = calloc(T,sizeof(int*));
    if(errno == ENOMEM) die("Couldn't get memory for g_ipt\n",errno);

#if (defined SSE || defined SSE2)
    gauge_field = calloc(4*VOLUMEPLUSRAND+1, sizeof(su3));
    if(errno == ENOMEM) die("Couldn't get memory for gauge_field\n",errno);
    g_gauge_field[0] = (su3*)(((unsigned int)(gauge_field)+ALIGN_BASE)&~ALIGN_BASE);
#else
    gauge_field = calloc(4*VOLUMEPLUSRAND, sizeof(su3));
    if(errno == ENOMEM) die("Couldn't get memory for gauge_field\n",errno);
    g_gauge_field[0] = gauge_field;
#endif
    idn = calloc(4*VOLUMEPLUSRAND, sizeof(int));
    if(errno == ENOMEM) die("Couldn't get memory for idn\n",errno);
    iup = calloc(4*VOLUMEPLUSRAND, sizeof(int));
    if(errno == ENOMEM) die("Couldn't get memory for iup\n",errno);
    ipt = calloc(T*LX*LY*LZ, sizeof(int));
    if(errno == ENOMEM) die("Couldn't get memory for ipt\n",errno);
    g_idn[0] = idn;
    g_iup[0] = iup;
    ipt_[0] = ipt;

    ipt__[0] = ipt_;
    g_ipt[0] = ipt__;
    for(i = 1; i < VOLUMEPLUSRAND; i++){
      g_idn[i] = g_idn[i-1]+4;
      g_iup[i] = g_iup[i-1]+4;
    }
    for(i = 1; i < VOLUMEPLUSRAND; i++){
      g_gauge_field[i] = g_gauge_field[i-1]+4;
    }
    for(i = 1; i < T*LX*LY; i++){
      ipt_[i] = ipt_[i-1]+LZ;
    }
    for(i = 1; i < T*LX; i++){
      ipt__[i] = ipt__[i-1]+LY;
    }
    for(i = 1; i < T; i++){
      g_ipt[i] = g_ipt[i-1]+LX;
    }
  }
}

void finalize(void){
  if(ini != 0){
    free(idn); 
    free(iup); 
    free(ipt);
    free(ipt_);
    free(ipt__);
    free(g_ipt);
    free(g_idn);
    free(g_iup);
    free(g_gauge_field);
    free(gauge_field);
  }
}
//===========================================================================
