/***************************************************************************
! plaquette.c
!
! Returns just the value of mean plaquette.
!   Code is based on GWC-Code function: 
!        --  void measure_gauge_action(int nstore)
!   written by Martin Hasenbusch <Martin.Hasenbusch@desy.de> Aug 2001
!
! Andre Sternbeck <andre.sternbeck@physik.hu-berlin.de
******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "global.h"
#include "su3.h"
#ifdef MPI
#include <mpi.h>
#endif

double plaquette(void)
{
  int     ix,ix1,ix2,mu1,mu2;
  static  su3 pr1,pr2; 
  su3     *v,*w;
  double  plaq; 
  double  sum;

  sum = 0.0 ;

  for ( ix=0; ix<VOLUME; ix++)
    {
      for ( mu1=0; mu1<3; mu1++ )
	{ 
	  ix1 = g_iup[ix][mu1];
	  for ( mu2=mu1+1; mu2<4; mu2++ ) 
	    { 
	      ix2 = g_iup[ix][mu2];
	      v = &g_gauge_field[ix][mu1];
	      w = &g_gauge_field[ix1][mu2];
	      _su3_times_su3(pr1,*v,*w);             // pr1 = v * w

	      v = &g_gauge_field[ix][mu2];
	      w = &g_gauge_field[ix2][mu1];
	      _su3_times_su3(pr2,*v,*w);             // pr2 = v * w

	      _trace_su3_times_su3d(plaq,pr1,pr2);   // plaq = ReTr(pr1*pr2^+)
	      sum = plaq+sum;
	    }
	}
    }

  /* normalization for plaquette VOLUME*6 */
#if defined MPI
  plaq=(sum)/3.0/(L*L*L*T*6.0);
  MPI_Allreduce(&plaq, &sum, 1, MPI_DOUBLE, MPI_SUM, g_cart_grid);
  plaq=sum;
#else
  plaq=(sum)/3.0/(VOLUME*6.0);
#endif
  return(plaq);
}

//=============================================================================
