#ifndef _SSE_H
#define _SSE_H
/*******************************************************************************
* $Id: sse.h,v 1.5 2003/05/23 14:32:32 urbach Exp $
*
* File sse.h
*
* Macros for Dirac spinors, SU(3) vectors and SU(3) matrices using
* inline assembly SSE and SSE2 instructions
*
* Needs gcc version 2.95.2 or later, and binutils snapshot 010122 or later
* if the SSE2 instructions are used
*
* Version: 1.1
* Author: Martin Luescher <luscher@mail.desy.de>
* Date: 17.03.2001
*
* with extensions by M.Hasenbusch
*
*******************************************************************************/

#if (defined SSE || defined SSE2 )
typedef struct
{
   int c0,c1,c2,c3;
} sse_int __attribute__ ((aligned (16)));

typedef struct
{
   float c0,c1,c2,c3;
} sse_float __attribute__ ((aligned (16)));

typedef struct
{
   double c0,c1;
} sse_double __attribute__ ((aligned (16)));

/*******************************************************************************
*
* Cache manipulation macros
*
*******************************************************************************/

#if defined P4

#define _prefetch_complex(addr) \
__asm__ __volatile__ ("prefetcht0 %0 \n\t" \
                      : \
                      : \
                      "m" (*(((char*)(addr)))))

#define _prefetch_nta_complex(addr) \
__asm__ __volatile__ ("prefetchnta %0 \n\t" \
                      : \
                      : \
                      "m" (*(((char*)(addr)))))

#define _prefetch_spinor(addr) \
__asm__ __volatile__ ("prefetcht0 %0 \n\t" \
                      "prefetcht0 %1" \
                      : \
                      : \
                      "m" (*(((char*)(((unsigned int)(addr))&~0x7f)))), \
                      "m" (*(((char*)(((unsigned int)(addr))&~0x7f))+128)))

#define _prefetch_nta_spinor(addr) \
__asm__ __volatile__ ("prefetchnta %0 \n\t" \
                      "prefetchnta %1" \
                      : \
                      : \
                      "m" (*(((char*)(((unsigned int)(addr))&~0x7f)))), \
                      "m" (*(((char*)(((unsigned int)(addr))&~0x7f))+128)))

#define _prefetch_su3(addr) \
__asm__ __volatile__ ("prefetcht0 %0 \n\t" \
                      "prefetcht0 %1" \
                      : \
                      : \
                      "m" (*(((char*)(((unsigned int)(addr))&~0x7f)))), \
                      "m" (*(((char*)(((unsigned int)(addr))&~0x7f))+128)))

#define _prefetch_mom(addr) \
__asm__ __volatile__ ("prefetchnta %0" \
                      : \
                      : \
                      "m" (*(((char*)(((unsigned int)(addr))&~0x7f))))) 
#else

#define _prefetch_complex(addr) \
__asm__ __volatile__ ("prefetcht0 %0 \n\t" \
                      "prefetcht0 %1 \n\t" \
                      "prefetcht0 %2 \n\t" \
                      "prefetcht0 %3 \n\t" \
                      : \
                      : \
                      "m" (*(((char*)(addr)))), \
                      "m" (*(((char*)(addr))+32)), \
                      "m" (*(((char*)(addr))+64)), \
                      "m" (*(((char*)(addr))+96)))

#define _prefetch_nta_complex(addr) \
__asm__ __volatile__ ("prefetchnta %0 \n\t" \
                      "prefetchnta %1 \n\t" \
                      "prefetchnta %2 \n\t" \
                      "prefetchnta %3 \n\t" \
                      : \
                      : \
                      "m" (*(((char*)(addr)))), \
                      "m" (*(((char*)(addr))+32)), \
                      "m" (*(((char*)(addr))+64)), \
                      "m" (*(((char*)(addr))+96)))


#define _prefetch_spinor(addr) \
__asm__ __volatile__ ("prefetcht0 %0 \n\t" \
                      "prefetcht0 %1 \n\t" \
                      "prefetcht0 %2 \n\t" \
                      "prefetcht0 %3 \n\t" \
                      "prefetcht0 %4 \n\t" \
                      "prefetcht0 %5" \
                      : \
                      : \
                      "m" (*(((char*)(addr)))), \
                      "m" (*(((char*)(addr))+32)), \
                      "m" (*(((char*)(addr))+64)), \
                      "m" (*(((char*)(addr))+96)), \
                      "m" (*(((char*)(addr))+128)), \
                      "m" (*(((char*)(addr))+160)))

#define _prefetch_nta_spinor(addr) \
__asm__ __volatile__ ("prefetchnta %0 \n\t" \
                      "prefetchnta %1 \n\t" \
                      "prefetchnta %2 \n\t" \
                      "prefetchnta %3 \n\t" \
                      "prefetchnta %4 \n\t" \
                      "prefetchnta %5" \
                      : \
                      : \
                      "m" (*(((char*)(addr)))), \
                      "m" (*(((char*)(addr))+32)), \
                      "m" (*(((char*)(addr))+64)), \
                      "m" (*(((char*)(addr))+96)), \
                      "m" (*(((char*)(addr))+128)), \
                      "m" (*(((char*)(addr))+160)))

#define _prefetch_su3(addr) \
__asm__ __volatile__ ("prefetcht0 %0  \n\t" \
                      "prefetcht0 %1  \n\t" \
                      "prefetcht0 %2  \n\t" \
                      "prefetcht0 %3  \n\t" \
                      "prefetcht0 %4" \
                      : \
                      : \
                      "m" (*(((char*)(((unsigned int)(addr))&~0x1f)))), \
                      "m" (*(((char*)(((unsigned int)(addr))&~0x1f))+32)), \
                      "m" (*(((char*)(((unsigned int)(addr))&~0x1f))+64)), \
                      "m" (*(((char*)(((unsigned int)(addr))&~0x1f))+96)), \
                      "m" (*(((char*)(((unsigned int)(addr))&~0x1f))+128)))

#define _prefetch_mom(addr) \
__asm__ __volatile__ ("prefetcht0 %0  \n\t" \
                      "prefetcht0 %1" \
                      : \
                      : \
                      "m" (*(((char*)(((unsigned int)(addr))&~0x1f)))), \
                      "m" (*(((char*)(((unsigned int)(addr))&~0x1f))+32)))

#endif

#if defined SSE2
#include"sse-extension/sse2.h"
#endif

#endif
#endif
